
//////////////////////////////////////////////////
//  DB Access Components
//  Copyright  1998-2024 Devart. All right reserved.
//  SQLGenerator Frame
//////////////////////////////////////////////////

{$IFNDEF CLR}

{$I Odac.inc}

unit OraSQLGeneratorFrame;
{$ENDIF}

interface

uses
{$IFDEF MSWINDOWS}
  Windows, Messages,
{$ENDIF}
  Graphics, Controls, Forms, Dialogs, StdCtrls, ExtCtrls, Buttons,
{$IFDEF FPC}
  LResources,
{$ENDIF}
  Classes, SysUtils,
  {$IFNDEF FPC}MemDS{$ELSE}MemDataSet{$ENDIF}, DBAccess, CRAccess,
  CRFrame, CRTabEditor, DASQLGeneratorFrame,
  Ora, DB, DASQLGenerator;

type
  TOraSQLGeneratorFrame = class(TDASQLGeneratorFrame)
    cbUseReturning: TCheckBox;
    cbUseRefresh: TCheckBox;
    cbFullRefresh: TCheckBox;
    btShowScript: TButton;
    gbSQLStyle: TGroupBox;
    rbDMLStyle: TRadioButton;
    rbSPStyle: TRadioButton;
    procedure btGenerateClick(Sender: TObject);
    procedure btShowScriptClick(Sender: TObject);
    procedure rbDMLStyleClick(Sender: TObject);
  protected
    FTablesInfo: TCRTablesInfo;
    FLastTablesInfoSQL: string;

    procedure DoActivate; override;

    procedure GenerateSelectFromAllFields; override;

    function GetTablesInfo: TCRTablesInfo; override;

    procedure UpdateControlsState; override;
    function GenerateStoredProcSQL(StatementType: TStatementType; GenerateScript: boolean): string;
    function GenerateSQLforUpdTable(TableInfo: TCRTableInfo;
      const KeyAndDataFields: TKeyAndDataFields;
      const StatementType: TStatementType;
      const ModifiedFieldsOnly: boolean): string; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

implementation

{$IFNDEF FPC}
{$IFDEF CLR}
{$R OraSQLGeneratorFrame.dfm}
{$ELSE}
{$R *.dfm}
{$ENDIF}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses
  DAQueryEditor, DAUpdateSQLEditor, DAUpdateSQLFrame, DAConsts, OraQueryEditor,
  DASQLFrame, CREditor, OraSPScriptEditor, DASQLComponentEditor, OraClasses;

constructor TOraSQLGeneratorFrame.Create(AOwner: TComponent);
begin
  inherited;

  FPromptlyServerRead := False;
  FParseTableNames := True;
  FTablesInfo := TCRTablesInfo.Create(TCRTableInfo);
end;

destructor TOraSQLGeneratorFrame.Destroy;
begin
  FTablesInfo.Free;
  inherited;
end;

procedure TOraSQLGeneratorFrame.DoActivate;
var
  FirstTime: boolean;
begin
  FirstTime := not FActivated;
  inherited;

  TCustomOraQuery(LocalDataSet).Options.ExtendedFieldsInfo := True;

  if FirstTime then begin
    if TCustomOraQuery(LocalDataSet).SQLLock.Text <> '' then
      cbUseRefresh.Checked := Pos('FOR', AnsiUpperCase(Trim(TCustomOraQuery(LocalDataSet).SQLLock.Text))) = 1;
    cbFullRefresh.Checked := TCustomOraQuery(LocalDataSet).Options.FullRefresh;
    cbUseReturning.Checked := TCustomOraQuery(LocalDataSet).DMLRefresh;
  end;
end;

procedure TOraSQLGeneratorFrame.GenerateSelectFromAllFields;
var
  IdentityField: TCRFieldDesc;
begin
  IdentityField := TDBAccessUtils.GetIdentityField(LocalDataSet);
  if (IdentityField <> nil) and (TablesInfo.IndexByName(SelectedTableName) >= 0) and
     (GenerateTableName(IdentityField) = SelectedTableName) then
    LocalDataSet.SQL.Text := 'SELECT ' + SelectedTableName + '.ROWID, ' +
      SelectedTableName + '.* FROM ' + SelectedTableName
  else
    inherited;
end;

function TOraSQLGeneratorFrame.GetTablesInfo: TCRTablesInfo;
var
  SQLInfo: TSQLInfo;
begin
  if FLastTablesInfoSQL <> LocalDataSet.FinalSQL then begin
    FTablesInfo.Clear;
    TDBAccessUtils.CheckIRecordSet(LocalDataSet);
    SQLInfo := TDBAccessUtils.GetSQLInfo(LocalDataSet);
    SQLInfo.ParseTablesInfo(LocalDataSet.FinalSQL, FTablesInfo);
    FLastTablesInfoSQL := LocalDataSet.FinalSQL;
  end;
  Result := FTablesInfo;
end;

procedure TOraSQLGeneratorFrame.UpdateControlsState;
var
  TableName: string;
  IsEnabled: boolean;
begin
  if FLockUpdateControlsState then
    Exit;

  inherited;

  TableName := UnqSelectedTableName;
  SetControlEnabled(btGetFields, TableName <> '');

  IsEnabled := btGenerate.Enabled;

  SetControlEnabled(cbUseReturning, IsEnabled);
  SetControlEnabled(cbUseRefresh, IsEnabled);
  SetControlEnabled(cbFullRefresh, IsEnabled);
  SetControlEnabled(rbDMLStyle, IsEnabled);
  SetControlEnabled(rbSPStyle, IsEnabled);
  SetControlEnabled(btShowScript, IsEnabled and rbSPStyle.Checked and
    (cbChecked(stInsert) or cbChecked(stUpdate) or cbChecked(stDelete)));
end;

function TOraSQLGeneratorFrame.GenerateStoredProcSQL(StatementType: TStatementType; GenerateScript: boolean): string;
const
  Prefix = 'p_';
  Blank = '  ';
var
  i, Append: integer;
  Table,
  St,
  FScript: string;

  function QuoteField(Name: string): string;
  begin
    if cbQuoteFields.Checked then
      Result := '"' + Name + '"'
    else
      Result := Name;
  end;

  function GetUpdate: string;
  var
    St: string;
    i, Append: integer;
    AValue: string;
    AReturning: string;
    AInto: string;
    ADataType: TFieldType;
  begin
    St := Blank + 'UPDATE ' + Table  + DALineSeparator + Blank + 'SET' + DALineSeparator;
    AReturning := '';
    AInto := '';
    Append := 0;
    for i := 0 to lbUpdateFields.Items.Count - 1 do
      if lbUpdateFields.Selected[i] then begin
        if Append <> 0 then
          St := St + ',' + DALineSeparator;

        ADataType := TFieldDef(LocalDataSet.FieldDefs.Find(lbUpdateFields.Items[i])).DataType;
        if (ADataType = ftOraBlob) or (ADataType = ftOraClob)
        then begin
          if ADataType = ftOraBlob then
            AValue := 'empty_blob()'
          else
            AValue := 'empty_clob()';
          if AReturning <> '' then
            AReturning := AReturning + ', ';
          if AInto <> '' then
            AInto := AInto + ', ';
          AReturning := AReturning + QuoteField(lbUpdateFields.Items[i]);
          AInto := AInto + Prefix + lbUpdateFields.Items[i];
        end
        else
          AValue := Prefix + lbUpdateFields.Items[i];
        St := St + Blank + '  ' + QuoteField(lbUpdateFields.Items[i]) + ' = ' +
          AValue;
        Inc(Append);
      end;

    St := St + DALineSeparator + Blank + 'WHERE' + DALineSeparator;

    Append := 0;
    for i := 0 to lbKeyFields.Items.Count - 1 do
      if lbKeyFields.Selected[i] then begin
        if Append <> 0 then
          St := St + ' and' + DALineSeparator;

        St := St + Blank + '  ' + QuoteField(lbKeyFields.Items[i]) + ' = ' +
          Prefix + 'OLD_' + lbKeyFields.Items[i];
        Inc(Append);
      end;
    if cbUseReturning.Checked then begin
      St := St + DALineSeparator + Blank + 'RETURNING' + DALineSeparator + Blank + '  ';

      Append := 0;
      for i := 0 to lbUpdateFields.Items.Count - 1 do
        if lbUpdateFields.Selected[i] then begin
          if Append <> 0 then
            St := St + ', ';

          St := St + QuoteField(lbUpdateFields.Items[i]);
          Inc(Append);
        end;

      St := St + DALineSeparator + Blank + 'INTO' + DALineSeparator + Blank + '  ';

      Append := 0;
      for i := 0 to lbUpdateFields.Items.Count - 1 do
        if lbUpdateFields.Selected[i] then begin
          if Append <> 0 then
            St := St + ', ';

          St := St + Prefix + lbUpdateFields.Items[i];
          Inc(Append);
        end;
    end else
    if (AReturning <> '') then
      St := St + DALineSeparator + Blank + 'RETURNING' + DALineSeparator + Blank + '  ' + AReturning +
        DALineSeparator + Blank + 'INTO' + DALineSeparator + Blank + '  ' + AInto;
    Result := St;
  end;

  function GetInsert: string;
  var
    St: string;
    i, Append: integer;
    ADataType: TFieldType;
    AReturning: string;
    AInto: string;
  begin
    St := Blank + 'INSERT INTO ' + Table + DALineSeparator + Blank + '  (';
    Append := 0;
    for i := 0 to lbUpdateFields.Items.Count - 1 do
      if lbUpdateFields.Selected[i] then begin
        if Append <> 0 then
          St := St + ', ';

        St := St + QuoteField(lbUpdateFields.Items[i]);
        Inc(Append);
      end;

    St := St + ')' + DALineSeparator + Blank + 'VALUES' + DALineSeparator + Blank + '  (';

    Append := 0;
    AReturning := '';
    AInto := '';
    for i := 0 to lbUpdateFields.Items.Count - 1 do
      if lbUpdateFields.Selected[i] then begin
        if Append <> 0 then
          St := St + ', ';

        ADataType := TFieldDef(LocalDataSet.FieldDefs.Find(lbUpdateFields.Items[i])).DataType;
        if (ADataType = ftOraBlob) or (ADataType = ftOraClob)
        then begin
          if ADataType = ftOraBlob then
            St := St + 'empty_blob()'
          else
            St := St + 'empty_clob()';
          if AReturning <> '' then
            AReturning := AReturning + ', ';
          if AInto <> '' then
            AInto := AInto + ', ';
          AReturning := AReturning + QuoteField(lbUpdateFields.Items[i]);
          AInto := AInto + Prefix + lbUpdateFields.Items[i];
        end
        else
          St := St + Prefix + lbUpdateFields.Items[i];

        Inc(Append);
      end;

    St := St + ')';

    if cbUseReturning.Checked then begin
      St := St + DALineSeparator + Blank + 'RETURNING' + DALineSeparator + Blank + '  ';

      Append := 0;
      for i := 0 to lbUpdateFields.Items.Count - 1 do
        if lbUpdateFields.Selected[i] then begin
          if Append <> 0 then
            St := St + ', ';

          St := St + QuoteField(lbUpdateFields.Items[i]);
          Inc(Append);
        end;

      St := St + DALineSeparator + Blank + 'INTO' + DALineSeparator + Blank + '  ';

      Append := 0;
      for i := 0 to lbUpdateFields.Items.Count - 1 do
        if lbUpdateFields.Selected[i] then begin
          if Append <> 0 then
            St := St + ', ';

          St := St + Prefix + lbUpdateFields.Items[i];
          Inc(Append);
        end;
    end else
    if (AReturning <> '') then
      St := St + DALineSeparator + Blank + 'RETURNING' + DALineSeparator + Blank + '  ' + AReturning +
        DALineSeparator + Blank + 'INTO' + DALineSeparator + Blank + '  ' + AInto;
    Result := St;
  end;

  function GetDelete: string;
  var
    St: string;
    i, Append: integer;
  begin

    St := Blank + 'DELETE FROM ' + Table + DALineSeparator + Blank + 'WHERE' + DALineSeparator;
    Append := 0;
    for i := 0 to lbKeyFields.Items.Count - 1 do
      if lbKeyFields.Selected[i] then begin
        if Append <> 0 then
          St := St + ' and' + DALineSeparator;

        St := St + Blank + '  ' + QuoteField(lbKeyFields.Items[i]) + ' = ' + Prefix;
        if lbKeyFields.Items[i] = 'ROWID' then
          St := St + 'OLD_' + lbKeyFields.Items[i]
        else
          St := St + lbKeyFields.Items[i];
        Inc(Append);
      end;
    Result := St;
  end;

  function GetType(FieldName: string): string;
  var
    FieldType: TFieldType;
  begin
    FieldType := TFieldDef(LocalDataSet.FieldDefs.Find(FieldName)).DataType;
    case Ord(FieldType) of
      Ord(ftString), Ord(ftFixedChar):
        Result := 'VARCHAR2';
      Ord(ftWideString), Ord(ftFixedWideChar):
        Result := 'NVARCHAR2';
      Ord(ftSmallint):
        Result := 'SMALLINT';
      Ord(ftInteger):
        Result := 'INTEGER';
      Ord(ftFloat), Ord(ftLargeInt), ftNumber, Ord(ftBCD):
        Result := 'NUMBER';
      Ord(ftDate):
        Result := 'DATE';
    {$IFDEF VER6P}
    {$IFNDEF FPC}
      Ord(ftTimeStamp),
    {$ENDIF}
    {$ENDIF}
      Ord(ftOraTimeStamp), Ord(ftDateTime):
        Result := 'TIMESTAMP';
      ftOraTimeStampTZ:
        Result := 'TIMESTAMP WITH TIME ZONE';
      ftOraTimeStampLTZ:
        Result := 'TIMESTAMP WITH LOCAL TIME ZONE';
      ftIntervalYM:
        Result := 'INTERVAL YEAR TO MONTH';
      ftIntervalDS:
        Result := 'INTERVAL DAY TO SECOND';
      else
        Result := ''
    end;
  end;

begin
  Table := Trim(cbTables.Text);
  i := Pos(' ', Table);
  if i > 0 then
    Table := Copy(Table, 1, i - 1);
  st := '';
  FScript := '';

  // INSERT
  if StatementType = stInsert then begin
    St := 'begin' + DALineSeparator + '  ' + Table + '_INSERT(';

    FScript := FScript + 'CREATE OR REPLACE' + DALineSeparator + 'PROCEDURE ' + Table + '_INSERT (' + DALineSeparator + '  ';
    Append := 0;
    for i := 0 to lbUpdateFields.Items.Count - 1 do
      if lbUpdateFields.Selected[i] then begin
        if Append <> 0 then begin
          St := St + ', ';
          FScript := FScript + ', ';
        end;

        St := St + ':' + lbUpdateFields.Items[i];
        FScript := FScript + 'p_' + lbUpdateFields.Items[i] + ' IN ';
        if cbUseReturning.Checked then
          FScript := FScript + 'OUT ';

        FScript := FScript + GetType(lbUpdateFields.Items[i]);
        Inc(Append);
      end;
    St := St + ');' + DALineSeparator + 'end;';
    if not GenerateScript then
      TCustomOraQuery(LocalDataSet).SQLInsert.Text := St;

    FScript := FScript + ')' + DALineSeparator + 'is' + DALineSeparator + 'begin' + DALineSeparator + GetInsert + ';' + DALineSeparator + 'end;' + DALineSeparator + '/' + DALineSeparator + DALineSeparator;
  end
  // UPDATE
  else
  if StatementType = stUpdate then begin
    St := 'begin' + DALineSeparator + '  ' + Table + '_UPDATE(';
    FScript := FScript + 'CREATE OR REPLACE' + DALineSeparator + 'PROCEDURE ' + Table + '_UPDATE (' + DALineSeparator + '  ';
    Append := 0;
  // KeyFields
    for i := 0 to lbKeyFields.Items.Count - 1 do
      if lbKeyFields.Selected[i] then begin
        if Append <> 0 then begin
          St := St + ', ';
          FScript := FScript + ', ';
        end;

        St := St + ':OLD_' + lbKeyFields.Items[i];
        FScript := FScript + 'p_OLD_' + lbKeyFields.Items[i] + ' IN ';

        if lbKeyFields.Items[i] = 'REC_USER' then
          beep;
        if lbKeyFields.Items[i] = 'ROWID' then
          FScript := FScript + 'VARCHAR2'
        else
          FScript := FScript + GetType(lbKeyFields.Items[i]);
        Inc(Append);
      end;
  // UpdateFields
    for i := 0 to lbUpdateFields.Items.Count - 1 do
      if lbUpdateFields.Selected[i] then begin
        if Append <> 0 then begin
          St := St + ', ';
          FScript := FScript + ', ';
        end;

        St := St + ':' + lbUpdateFields.Items[i];
        FScript := FScript + 'p_' + lbUpdateFields.Items[i] + ' IN ';
        if cbUseReturning.Checked then
          FScript := FScript + 'OUT ';

        FScript := FScript + GetType(lbUpdateFields.Items[i]);

        Inc(Append);
      end;
    St := St + ');' + DALineSeparator + 'end;';
    if not GenerateScript then
      TCustomOraQuery(LocalDataSet).SQLUpdate.Text := St;

    FScript := FScript + ')' + DALineSeparator + 'is' + DALineSeparator + 'begin' + DALineSeparator + GetUpdate + ';' + DALineSeparator + 'end;' + DALineSeparator + '/' + DALineSeparator + DALineSeparator;
  end

  // DELETE
  else
  if StatementType = stDelete then begin
    St := 'begin' + DALineSeparator + '  ' + Table + '_DELETE(';
    FScript := FScript + 'CREATE OR REPLACE' + DALineSeparator + 'PROCEDURE ' + Table + '_DELETE (' + DALineSeparator + '  ';
    Append := 0;
    for i := 0 to lbKeyFields.Items.Count - 1 do
      if lbKeyFields.Selected[i] then begin
        if Append <> 0 then begin
          St := St + ', ';
          FScript := FScript + ', ';
        end;

        St := St + ':' + lbKeyFields.Items[i];
        FScript := FScript + 'p_' + lbKeyFields.Items[i] + ' IN ';

        if lbKeyFields.Items[i] = 'ROWID' then
          FScript := FScript + 'VARCHAR2'
        else
          FScript := FScript + GetType(lbKeyFields.Items[i]);

        Inc(Append);
      end;
    St := St + ');' + DALineSeparator + 'end;';
    if not GenerateScript then
      TCustomOraQuery(LocalDataSet).SQLDelete.Text := St;

    FScript := FScript + ')' + DALineSeparator + 'is' + DALineSeparator + 'begin' + DALineSeparator + GetDelete + ';' + DALineSeparator + 'end;' + DALineSeparator + '/' + DALineSeparator + DALineSeparator;
  end;
  if GenerateScript then
    Result := FScript
  else
    Result := st;
end;

function TOraSQLGeneratorFrame.GenerateSQLforUpdTable(
  TableInfo: TCRTableInfo; const KeyAndDataFields: TKeyAndDataFields;
  const StatementType: TStatementType;
  const ModifiedFieldsOnly: boolean): string;
begin
  if (StatementType = stLock) and cbUseRefresh.Checked then
    Result := 'FOR UPDATE NOWAIT'
  else
  if rbDMLStyle.Checked or not (StatementType in [stInsert, stUpdate, stDelete]) then
    // DML
    Result := inherited GenerateSQLforUpdTable(TableInfo, KeyAndDataFields, StatementType, ModifiedFieldsOnly)
  else
    // StoredProc
    Result := GenerateStoredProcSQL(StatementType, False);
end;

procedure TOraSQLGeneratorFrame.btGenerateClick(Sender: TObject);
var
  OldFullRefresh,
  OldDMLRefresh: boolean;
//  KeyAndDataFields: TKeyAndDataFields;
begin
  OldDMLRefresh := TOraDataSet(LocalDataSet).DMLRefresh;
  OldFullRefresh := TOraDataSet(LocalDataSet).Options.FullRefresh;
  try
    TOraDataSet(LocalDataSet).DMLRefresh := cbUseReturning.Checked;
    TOraDataSet(LocalDataSet).Options.FullRefresh := cbFullRefresh.Checked;
    inherited;
  finally
    if LocalDataSet <> nil then begin
      TOraDataSet(LocalDataSet).Options.FullRefresh := OldFullRefresh;
      TOraDataSet(LocalDataSet).DMLRefresh := OldDMLRefresh;
    end;
  end;
end;

procedure TOraSQLGeneratorFrame.btShowScriptClick(Sender: TObject);
var
  OldSQL: string;
  OldActive: boolean;
  Script: string;
  st: TStatementType;
begin
  GetFields(True);
  OldSQL := LocalDataSet.SQL.Text;
  OldActive := LocalDataSet.Active;
  try
    if LocalDataSet.SQL.Text <> FLastGeneratedSQL then
      LocalDataSet.SQL.Text := FLastGeneratedSQL;
    OpenDataSet(True);
    Assert(Editor is TOraQueryEditorForm);
    Script := '';
    for st := Low(TStatementType) to High(TStatementType) do
      if (st in [stInsert, stUpdate, stDelete]) and cbChecked(st) then
        Script := Script + GenerateStoredProcSQL(st, True);
    with TOraSPScriptEditor.Create(Editor, Editor.DADesignUtilsClass) do
      try
        SQL := '-- Generated by ODAC' + DALineSeparator + DALineSeparator + Script;
        ShowModal;
      finally
        Free;
      end;
  finally
    if not OldActive then
      LocalDataSet.SQL.Text := OldSQL;
  end;
end;

procedure TOraSQLGeneratorFrame.rbDMLStyleClick(Sender: TObject);
begin
  UpdateControlsState;
end;

end.
