
//////////////////////////////////////////////////
//  Oracle Data Access Components
//  Copyright  1998-2024 Devart. All right reserved.
//  Oracle Error Handler
//////////////////////////////////////////////////


{$IFNDEF CLR}

{$I Odac.inc}
unit OraErrHand;
{$ENDIF}
interface

uses
{$IFDEF MSWINDOWS}
  Windows, Messages,
{$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
{$IFDEF FPC}
  LCLType,
{$ENDIF}
  CRTypes, Ora;

type
  TOnOraErrorEvent = procedure (Sender: TObject; E: Exception; ErrorCode: integer;
    const ConstraintName: string; var Msg: string) of object;

  TOnError1Event = procedure (Sender: TObject; E: Exception; ErrorCode: integer;
    const ConstraintName: string; var Msg: string; var Handled: boolean) of object;

{$IFDEF VER16P}
  [ComponentPlatformsAttribute({$IFDEF STD}pidDevartWinPlatforms{$ELSE}pidDevartAllPlatforms{$ENDIF})]
{$ENDIF}
  TOraErrorHandler = class(TComponent)
  private
    FQuery: TOraQuery;
    FTableName: string;
    FDebug: boolean;
    FStreamedActive: boolean;
    FOnError: TOnOraErrorEvent;
    FOnError1: TOnError1Event; // DEBUG

    function GetSession: TOraSession;
    procedure SetSession(Value: TOraSession);
    function GetActive: boolean;
    procedure SetActive(Value: boolean);
    procedure SetTableName(Value: string);

  protected
    FDesignCreate: boolean;

    procedure Loaded; override;

    procedure OnException(Sender: TObject; E: Exception);

    function FindError(SenderName: string; ErrorClass: string; ErrorCode: integer;
      const ConstraintName: string; const FullConstraintName: string; var Msg: string): boolean;

  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;

    procedure Open;
    procedure Close;

    property Query: TOraQuery read FQuery;

    property OnError1: TOnError1Event read FOnError1 write FOnError1;

  published
    property Session: TOraSession read GetSession write SetSession;
    property TableName: string read FTableName write SetTableName;
    property Active: boolean read GetActive write SetActive default False;
    property Debug: boolean read FDebug write FDebug default False;
    property OnError: TOnOraErrorEvent read FOnError write FOnError;
  end;

  TOraErrorHandlerUtils = class
  public
    class procedure SetDesignCreate(Obj: TOraErrorHandler; Value: boolean);
    class function GetDesignCreate(Obj: TOraErrorHandler): boolean;
  end;

implementation

uses
  OraError;

{ TOraErrorHandler }

constructor TOraErrorHandler.Create(Owner: TComponent);
begin
  inherited Create(Owner);

  FDebug := False;
  FOnError := nil;
  FTableName := '';

  Application.OnException := OnException;
  FQuery := TOraQuery.Create(nil);
  FStreamedActive:= False;

  FDesignCreate := csDesigning in ComponentState;
end;

destructor TOraErrorHandler.Destroy;
begin
  FQuery.Free;
  Application.OnException := nil;

  inherited;
end;

procedure TOraErrorHandler.Open;
begin
  Active := True;
end;

procedure TOraErrorHandler.Close;
begin
  Active := False;
end;

procedure TOraErrorHandler.Loaded;
begin
  inherited;

  FDesignCreate := False;

  try
    if FStreamedActive then
      Active := True;
  except
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

function TOraErrorHandler.FindError(SenderName: string; ErrorClass: string; ErrorCode: integer;
  const ConstraintName: string; const FullConstraintName: string; var Msg: string): boolean;
var
  Code: integer;
  Cons: string;
begin
  if Active then begin
    FQuery.First;

    Result := False;
    while not FQuery.EOF do begin
      Code := FQuery.FieldByName('ErrorCode').AsInteger;
      Cons := UpperCase(FQuery.FieldByName('Constraint').AsString);
      if Code > ErrorCode then
        break;

      if Code = ErrorCode then
        if (Cons = '') or (Cons = UpperCase(ConstraintName)) or (Cons = UpperCase(FullConstraintName)) then begin
          Result:= True;
          break;
        end;
      FQuery.Next;
    end;

    if Result then
      Msg := FQuery.FieldByName('Message').AsString
    else
      Msg := '';
  end
  else
    Result := False;
end;

procedure TOraErrorHandler.OnException(Sender: TObject; E: Exception);
var
  Msg: string;
  SenderName: string;
  ErrorClass: string;
  ErrorCode: integer;
  Constraint: string;
  FullConstraint: string;
  DebugSt: string;
  St: string;
  i, j: integer;
  Handled: boolean;
begin
  ErrorClass := E.ClassName;

// Sender Name
  SenderName := '';
  try  // Sender may be destroyed WAR
    if E is EOraError then
      if EOraError(E).Sender <> nil then begin
        SenderName := EOraError(E).Sender.Name;
        Sender := EOraError(E).Sender;
      end;
    if Sender = nil then
      Sender := self;
    if (SenderName = '') and (Sender is TComponent) then
      SenderName := TComponent(Sender).Name;
  except
  {$IFNDEF CLR}
    on EAccessViolation do  // WAR
    else
  {$ENDIF}
      raise;
  end;

// Error Code
  if E is EOraError then
    ErrorCode := EOraError(E).ErrorCode
  else
    ErrorCode := 0;

// Constraint
  if (E is EOraError) and
    (ErrorCode = 0001) or  // UNIQUE
    (ErrorCode = 2290) or  // CHECK
    (ErrorCode = 2291) or  // FOREG
    (ErrorCode = 2292)
  then begin
    St := E.Message;
    i := Pos('(', St);
    if i > 0 then
      St := Copy(St, i + 1, Length(St) - i - 1);

    j := Pos(')', St);
    FullConstraint := Copy(St, 1, j - 1);

    i := Pos('.', St);
    if (i > 0) and (i < j) then begin
      St := Copy(St, i + 1, Length(St) - i - 1);
      j := j - i;
    end;
    SetLength(St, j - 1);

    Constraint := St;
  end;

  if Debug then begin
    if ErrorCode <> 0 then
      St := IntToStr(ErrorCode)
    else
      St := '';

    DebugSt :=
      'Sender Name:   ' + SenderName + #13 +
      'Except. Class:   ' + ErrorClass + #13 +
      'Error Code:       ' + St + #13 +
      'Constraint:        ' + Constraint + #13 + #13
  end
  else
    DebugSt := '';

  if not Assigned(FOnError1) then begin
  // original behavior
    if not FindError(SenderName, ErrorClass, ErrorCode, Constraint, FullConstraint, Msg) then
      if Assigned(FOnError) then
        FOnError(Sender, E, ErrorCode, Constraint, Msg);

    if Msg <> '' then begin
      if Debug then
        DebugSt := DebugSt + 'Original Msg:    ' + E.Message + #13 + #13;
    end
    else
      Msg := E.Message;

  {$IFNDEF CLR}
    Application.MessageBox(PChar(DebugSt + Msg), PChar(Application.Title), MB_OK + MB_ICONSTOP);
  {$ENDIF}
  end
  else begin
  // DEBUG
    FindError(SenderName, ErrorClass, ErrorCode, Constraint, FullConstraint, Msg);

    Handled := False;

    if Assigned(FOnError1) then
      FOnError1(Sender, E, ErrorCode, Constraint, Msg, Handled);

    if Msg <> '' then begin
      if Debug then
        DebugSt := DebugSt + 'Original Msg:    ' + E.Message + #13 + #13;
    end
    else
      Msg := E.Message;

    if not Handled then
      Application.MessageBox(PChar(DebugSt + Msg), PChar(Application.Title), MB_OK + MB_ICONSTOP);
  end;
end;

function TOraErrorHandler.GetSession: TOraSession;
begin
  Result := FQuery.Session;
end;

procedure TOraErrorHandler.SetSession(Value: TOraSession);
begin
  if Value <> FQuery.Session then begin
    Close;
    FQuery.Session := Value;
  end;
end;

function TOraErrorHandler.GetActive: boolean;
begin
  Result := FQuery.Active;
end;

procedure TOraErrorHandler.SetActive(Value: boolean);
begin
  if Value and (FTableName = '') then
    raise Exception.Create('TableName must be defined');

  if csReading in ComponentState then begin
    if Value then
       FStreamedActive := True;
  end
  else
    try
      FQuery.Active := Value;
    except
      on E:TObject do begin
        if E is EOraError then
          EOraError(E).Sender := Self;
        raise;
      end;
    end; // try
end;

procedure TOraErrorHandler.SetTableName(Value: string);
begin
  FTableName := Value;
  if FTableName <> '' then begin
    FQuery.SQL.Clear;
    FQuery.SQL.Add('SELECT E.*, E.RowId FROM ' + FTableName + ' E ' +
      'ORDER BY ErrorCode, Constraint');
  end;
end;

class procedure TOraErrorHandlerUtils.SetDesignCreate(Obj: TOraErrorHandler; Value: boolean);
begin
  Obj.FDesignCreate := Value;
end;

class function TOraErrorHandlerUtils.GetDesignCreate(Obj: TOraErrorHandler): boolean;
begin
  Result := Obj.FDesignCreate;
end;

initialization
{$IFDEF VER16P}
  GroupDescendentsWith(TOraErrorHandler, TControl);
{$ENDIF}

end.
